import py
import sys

class AppTestUnicodeData:
    spaceconfig = dict(usemodules=('unicodedata',))

    def test_hangul_syllables(self):
        import unicodedata
        # Test all leading, vowel and trailing jamo
        # but not every combination of them.
        for code, name in ((0xAC00, 'HANGUL SYLLABLE GA'),
                           (0xAE69, 'HANGUL SYLLABLE GGAEG'),
                           (0xB0D2, 'HANGUL SYLLABLE NYAGG'),
                           (0xB33B, 'HANGUL SYLLABLE DYAEGS'),
                           (0xB5A4, 'HANGUL SYLLABLE DDEON'),
                           (0xB80D, 'HANGUL SYLLABLE RENJ'),
                           (0xBA76, 'HANGUL SYLLABLE MYEONH'),
                           (0xBCDF, 'HANGUL SYLLABLE BYED'),
                           (0xBF48, 'HANGUL SYLLABLE BBOL'),
                           (0xC1B1, 'HANGUL SYLLABLE SWALG'),
                           (0xC41A, 'HANGUL SYLLABLE SSWAELM'),
                           (0xC683, 'HANGUL SYLLABLE OELB'),
                           (0xC8EC, 'HANGUL SYLLABLE JYOLS'),
                           (0xCB55, 'HANGUL SYLLABLE JJULT'),
                           (0xCDBE, 'HANGUL SYLLABLE CWEOLP'),
                           (0xD027, 'HANGUL SYLLABLE KWELH'),
                           (0xD290, 'HANGUL SYLLABLE TWIM'),
                           (0xD4F9, 'HANGUL SYLLABLE PYUB'),
                           (0xD762, 'HANGUL SYLLABLE HEUBS'),
                           (0xAE27, 'HANGUL SYLLABLE GYIS'),
                           (0xB090, 'HANGUL SYLLABLE GGISS'),
                           (0xB0AD, 'HANGUL SYLLABLE NANG'),
                           (0xB316, 'HANGUL SYLLABLE DAEJ'),
                           (0xB57F, 'HANGUL SYLLABLE DDYAC'),
                           (0xB7E8, 'HANGUL SYLLABLE RYAEK'),
                           (0xBA51, 'HANGUL SYLLABLE MEOT'),
                           (0xBCBA, 'HANGUL SYLLABLE BEP'),
                           (0xBF23, 'HANGUL SYLLABLE BBYEOH'),
                           (0xD7A3, 'HANGUL SYLLABLE HIH')):
            assert unicodedata.name(unichr(code)) == name
            assert unicodedata.lookup(name) == unichr(code)
        # Test outside the range
        raises(ValueError, unicodedata.name, unichr(0xAC00 - 1))
        raises(ValueError, unicodedata.name, unichr(0xD7A3 + 1))

    def test_cjk(self):
        import sys
        import unicodedata
        cases = ((0x3400, 0x4DB5),
                 (0x4E00, 0x9FA5))
        if unicodedata.unidata_version >= "5":    # don't know the exact limit
            cases = ((0x3400, 0x4DB5),
                     (0x4E00, 0x9FCB),
                     (0x20000, 0x2A6D6),
                     (0x2A700, 0x2B734))
        elif unicodedata.unidata_version >= "4.1":
            cases = ((0x3400, 0x4DB5),
                     (0x4E00, 0x9FBB),
                     (0x20000, 0x2A6D6))
        for first, last in cases:
            # Test at and inside the boundary
            for i in (first, first + 1, last - 1, last):
                charname = 'CJK UNIFIED IDEOGRAPH-%X'%i
                char = ('\\U%08X' % i).decode('unicode-escape')
                assert unicodedata.name(char) == charname
                assert unicodedata.lookup(charname) == char
            # Test outside the boundary
            for i in first - 1, last + 1:
                charname = 'CJK UNIFIED IDEOGRAPH-%X'%i
                char = ('\\U%08X' % i).decode('unicode-escape')
                try:
                    unicodedata.name(char)
                except ValueError as e:
                    assert e.message == 'no such name'
                raises(KeyError, unicodedata.lookup, charname)

    def test_bug_1704793(self): # from CPython
        import unicodedata
        assert unicodedata.lookup("GOTHIC LETTER FAIHU") == u'\U00010346'

    def test_no_alias_python2(self):
        import unicodedata
        with raises(KeyError):
            assert unicodedata.lookup("LATIN SMALL LETTER GHA")
        assert unicodedata.name(u"\u01a3") == "LATIN SMALL LETTER OI"

    def test_normalize_bad_argcount(self):
        import unicodedata
        raises(TypeError, unicodedata.normalize, 'x')

    def test_normalize_nonunicode(self):
        import unicodedata
        exc_info = raises(TypeError, unicodedata.normalize, 'NFC', 'x')
        assert str(exc_info.value).endswith('must be unicode, not str')

    @py.test.mark.skipif("sys.maxunicode < 0x10ffff")
    def test_normalize_wide(self):
        import unicodedata
        assert unicodedata.normalize('NFC', u'\U000110a5\U000110ba') == u'\U000110ab'

    def test_linebreaks(self):
        linebreaks = (0x0a, 0x0b, 0x0c, 0x0d, 0x85,
                      0x1c, 0x1d, 0x1e, 0x2028, 0x2029)
        for i in linebreaks:
            for j in range(-2, 3):
                lines = (unichr(i + j) + u'A').splitlines()
                if i + j in linebreaks:
                    assert len(lines) == 2
                else:
                    assert len(lines) == 1

    def test_mirrored(self):
        import unicodedata
        # For no reason, unicodedata.mirrored() returns an int, not a bool
        assert repr(unicodedata.mirrored(u' ')) == '0'

    def test_bidirectional_not_one_character(self):
        import unicodedata
        exc_info = raises(TypeError, unicodedata.bidirectional, u'xx')
        assert str(exc_info.value) == 'need a single Unicode character as parameter'

    def test_bidirectional_not_one_character(self):
        import unicodedata
        exc_info = raises(TypeError, unicodedata.bidirectional, 'x')
        assert str(exc_info.value).endswith('must be unicode, not str')



from rpython.tool.uid import uid

# Support for explicit specialization: in code using global constants
# that are instances of SpecTag, code paths are not merged when
# the same variable holds a different SpecTag instance.

class SpecTag(object):
    __slots__ = ()

    def __repr__(self):
        return '%s(0x%x)' % (self.__class__.__name__, uid(self))

    def _freeze_(self):
        return True


# 'for' iteration over iterables wrapped in an instance
# of unrolling_iterable will be unrolled by the flow space,
# like in:
#     names = unrolling_iterable(['a', 'b', 'c'])
#     def f(x):
#         for name in names:
#             setattr(x, name, 0)

class unrolling_iterable(SpecTag):

    def __init__(self, iterable):
        self._items = list(iterable)
        self._head = _unroller(self._items)

    def __iter__(self):
        return iter(self._items)

    def get_unroller(self):
        return self._head


class _unroller(SpecTag):

    def __init__(self, items, i=0):
        self._items = items
        self._i = i
        self._next = None

    def step(self):
        v = self._items[self._i]
        if self._next is None:
            self._next = _unroller(self._items, self._i+1)
        return v, self._next



from rpython.rtyper.rmodel import Repr
from rpython.rtyper.lltypesystem import lltype, llmemory

from rpython.tool.pairtype import pairtype, extendabletype, pair

UNKNOWN = object()

class GCRefRepr(Repr):
    lowleveltype = llmemory.GCREF

    @staticmethod
    def make(r_base, cache):
        try:
            return cache[r_base]
        except KeyError:
            res = cache[r_base] = GCRefRepr(r_base)
            return res


    def __init__(self, r_base):
        self.r_base = r_base
        self._ll_eq_func = UNKNOWN
        self._ll_hash_func = UNKNOWN
        if hasattr(r_base, 'll_str'):
            ll_base_str = r_base.ll_str
            def ll_str(ptr):
                return ll_base_str(lltype.cast_opaque_ptr(r_base.lowleveltype, ptr))
            self.ll_str = ll_str

    def convert_const(self, x):
        return lltype.cast_opaque_ptr(llmemory.GCREF, self.r_base.convert_const(x))

    def get_ll_eq_function(self):
        if self._ll_eq_func is UNKNOWN:
            ll_base_eq_function = self.r_base.get_ll_eq_function()
            if ll_base_eq_function is None:
                ll_eq_func = None
            else:
                def ll_eq_func(ptr1, ptr2):
                    ptr1 = lltype.cast_opaque_ptr(self.r_base.lowleveltype, ptr1)
                    ptr2 = lltype.cast_opaque_ptr(self.r_base.lowleveltype, ptr2)
                    return ll_base_eq_function(ptr1, ptr2)
            self._ll_eq_func = ll_eq_func
        return self._ll_eq_func

    def get_ll_hash_function(self):
        if self._ll_hash_func is UNKNOWN:
            ll_base_hash_function = self.r_base.get_ll_hash_function()
            if ll_base_hash_function is None:
                ll_hash_func = None
            else:
                def ll_hash_func(ptr):
                    ptr = lltype.cast_opaque_ptr(self.r_base.lowleveltype, ptr)
                    return ll_base_hash_function(ptr)
            self._ll_hash_func = ll_hash_func
        return self._ll_hash_func

    def get_ll_dummyval_obj(self, rtyper, s_value):
        return DummyValueBuilderGCRef(rtyper)

class __extend__(pairtype(GCRefRepr, Repr)):
    def convert_from_to((r_from, r_to), v, llops):
        if isinstance(r_to.lowleveltype, lltype.Ptr) and r_to.lowleveltype.TO._gckind == 'gc':
            return llops.genop('cast_opaque_ptr', [v], r_to.lowleveltype)
        return NotImplemented

class __extend__(pairtype(Repr, GCRefRepr)):
    def convert_from_to((r_from, r_to), v, llops):
        if r_from != r_to.r_base:
            v = pair(r_from, r_to.r_base).convert_from_to(v, llops)
        return llops.genop('cast_opaque_ptr', [v], r_to.lowleveltype)


class DummyValueBuilderGCRef(object):

    def __init__(self, rtyper):
        self.rtyper = rtyper

    def _freeze_(self):
        return True

    def __hash__(self):
        return hash(llmemory.GCREF)

    def __eq__(self, other):
        return (isinstance(other, DummyValueBuilderGCRef) and
                self.rtyper is other.rtyper)

    def __ne__(self, other):
        return not (self == other)

    @property
    def ll_dummy_value(self):
        try:
            return self.rtyper.cache_dummy_values[llmemory.GCREF]
        except KeyError:
            from rpython.rtyper import rclass
            from rpython.rtyper.rmodel import DummyValueBuilder
            rinstbase = rclass.getinstancerepr(self.rtyper, None)
            TYPE = rinstbase.lowleveltype
            val = DummyValueBuilder(self.rtyper, TYPE.TO).ll_dummy_value
            p = lltype.cast_opaque_ptr(llmemory.GCREF, val)
            self.rtyper.cache_dummy_values[llmemory.GCREF] = p
            return p





from rpython.rtyper.lltypesystem import lltype
from rpython.rtyper import rclass

def is_immutable_struct(S):
    return isinstance(S, lltype.GcStruct) and S._hints.get('immutable', False)

def has_gcstruct_a_vtable(GCSTRUCT):
    if not isinstance(GCSTRUCT, lltype.GcStruct):
        return False
    if GCSTRUCT is rclass.OBJECT:
        return False
    while not GCSTRUCT._hints.get('typeptr'):
        _, GCSTRUCT = GCSTRUCT._first_struct()
        if GCSTRUCT is None:
            return False
    return True

def get_vtable_for_gcstruct(gccache, GCSTRUCT):
    # xxx hack: from a GcStruct representing an instance's
    # lowleveltype, return the corresponding vtable pointer.
    # Returns None if the GcStruct does not belong to an instance.
    if not isinstance(GCSTRUCT, lltype.GcStruct):
        return lltype.nullptr(rclass.OBJECT_VTABLE)
    if not has_gcstruct_a_vtable(GCSTRUCT):
        return lltype.nullptr(rclass.OBJECT_VTABLE)
    setup_cache_gcstruct2vtable(gccache)
    try:
        return gccache._cache_gcstruct2vtable[GCSTRUCT]
    except KeyError:
        return testing_gcstruct2vtable[GCSTRUCT]

def setup_cache_gcstruct2vtable(gccache):
    if not hasattr(gccache, '_cache_gcstruct2vtable'):
        cache = {}
        if gccache.rtyper:
            for rinstance in gccache.rtyper.instance_reprs.values():
                cache[rinstance.lowleveltype.TO] = rinstance.rclass.getvtable()
        gccache._cache_gcstruct2vtable = cache

def set_testing_vtable_for_gcstruct(GCSTRUCT, vtable, name):
    # only for tests that need to register the vtable of their malloc'ed
    # structures in case they are GcStruct inheriting from OBJECT.
    vtable.name = rclass.alloc_array_name(name)
    testing_gcstruct2vtable[GCSTRUCT] = vtable

testing_gcstruct2vtable = {}

# ____________________________________________________________


def all_fielddescrs(gccache, STRUCT, only_gc=False, res=None,
                    get_field_descr=None):
    from rpython.jit.backend.llsupport import descr

    if get_field_descr is None:
        get_field_descr = descr.get_field_descr
    if res is None:
        res = []
    # order is not relevant, except for tests
    for name in STRUCT._names:
        FIELD = getattr(STRUCT, name)
        if FIELD is lltype.Void:
            continue
        if name.startswith('c__pad'):
            continue
        if name == 'typeptr':
            continue # dealt otherwise
        elif isinstance(FIELD, lltype.Struct):
            all_fielddescrs(gccache, FIELD, only_gc, res, get_field_descr)
        elif (not only_gc) or (isinstance(FIELD, lltype.Ptr) and FIELD._needsgc()):
            res.append(get_field_descr(gccache, STRUCT, name))
    return res

def all_interiorfielddescrs(gccache, ARRAY, get_field_descr=None):
    from rpython.jit.backend.llsupport import descr
    from rpython.jit.codewriter.effectinfo import UnsupportedFieldExc

    if get_field_descr is None:
        get_field_descr = descr.get_field_descr
    # order is not relevant, except for tests
    STRUCT = ARRAY.OF
    res = []
    for name in STRUCT._names:
        FIELD = getattr(STRUCT, name)
        if FIELD is lltype.Void:
            continue
        if name == 'typeptr':
            continue # dealt otherwise
        elif isinstance(FIELD, lltype.Struct):
            raise UnsupportedFieldExc("unexpected array(struct(struct))")
        res.append(get_field_descr(gccache, ARRAY, name))
    return res

def gc_fielddescrs(gccache, STRUCT):
    return all_fielddescrs(gccache, STRUCT, True)

def get_fielddescr_index_in(STRUCT, fieldname, cur_index=0):
    for name in STRUCT._names:
        FIELD = getattr(STRUCT, name)
        if FIELD is lltype.Void:
            continue
        if name == 'typeptr':
            continue # dealt otherwise
        elif isinstance(FIELD, lltype.Struct):
            r = get_fielddescr_index_in(FIELD, fieldname, cur_index)
            if r >= 0:
                return r
            cur_index += -r - 1
            continue
        elif name == fieldname:
            return cur_index
        cur_index += 1
    return -cur_index - 1 # not found
